@@ -1,4 +1,4 @@
-# (C) Copyright 1996- ECMWF.
+﻿# (C) Copyright 1996- ECMWF.
 #
 # This software is licensed under the terms of the Apache Licence Version 2.0
 # which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
@@ -18,7 +18,7 @@
     * Humidex
     * Apparent Temperature
     * Wind Chill
-    * Net Effective Temperature
+    * Normal Effective Temperature (NET)
 
     In support of the above indexes, it also calculates:
     * Solar Declination Angle
@@ -36,8 +36,8 @@ import numpy as np
 
 from .helpers import (
     __wrap,
-    farenheit_to_celcius,
-    kelvin_to_farenheit,
+    fahrenheit_to_celsius,
+    kelvin_to_fahrenheit,
     kPa_to_hPa,
     to_julian_date,
     to_radians,
@@ -82,8 +82,8 @@ def calculate_relative_humidity_percent(t2m, td):
     t2m = __wrap(t2m)
     td = __wrap(td)
 
-    t2m = kelvin_to_celcius(t2m)
-    td = kelvin_to_celcius(td)
+    t2m = kelvin_to_celsius(t2m)
+    td = kelvin_to_celsius(td)
 
     # saturated vapour pressure
     es = 6.11 * 10.0 ** (7.5 * t2m / (237.3 + t2m))
@@ -161,159 +161,99 @@ def calculate_cos_solar_zenith_angle(h, lat, lon, y, m, d):
     return np.clip(csza, 0, None)
 
 
-def calculate_cos_solar_zenith_angle_integrated(lat, lon, y, m, d, h, base, step):
+def calculate_cos_solar_zenith_angle_integrated(
+    lat, lon, y, m, d, h, tbegin, tend, intervals_per_hour=1, integration_order=3
+):
     """
-    calculate solar zenith angle
+    calculate average of solar zenith angle based on numerical integration using 3 point gauss integration rule
     :param lat: (int array) latitude [degrees]
     :param lon: (int array) longitude [degrees]
     :param y: year [int]
     :param m: month [int]
     :param d: day [int]
     :param h: hour [int]
-    :param base: base time of forecast enum [0,6,12,18]
-    :param step: step interval of forecast enum [0,1,3,6,..]
+    :param tbegin: offset in hours from forecast time to begin of time interval for integration [int]
+    :param tend:  offset in hours from forecast time to end of time interval for integration [int]
+    :param integration order:  order of gauss integration [int] valid = (1, 2, 3, 4)
+    :param intervals_per_hour:  number of time intregrations per hour [int]
 
     https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1002/2015GL066868
 
-    returns cosine of the solar zenith angle [degrees]
+    This uses Gaussian numerical integration. See https://en.wikipedia.org/wiki/Gaussian_quadrature
+
+    returns average of cosine of the solar zenith angle during interval [degrees]
     """
 
-    # OLD CODE
-    # maxx = 0
-    # base_offset = 0
-    # h_offset = 0
-    # # step_offset = 0
-    # if base == 0:
-    #     base_offset = 0
-    # if base == 6:
-    #     base_offset = 594
-    # if base == 18:
-    #     base_offset = 1806
-    # if base == 12:
-    #     base_offset = 1212  # 1200+12
-
-    # if step == 1:
-    #     h_offset = 0.5
-    #     maxx = 32.0
-    # if step == 3:
-    #     h_offset = 1.5
-    #     maxx = 14.0
-    # if step == 6:
-    #     h_offset = 3
-    #     maxx = 100.0
-
-    # hh = h - base_offset - h_offset
-
-    # POTENTIAL BUG in OLD CODE ABOVE -- POSSIBLE FIX BELOW -- Needs checking
-    # This ensures we can integrate but doesn't use hours offsets
-    hh = h
-    maxx = 100.0
-
-    # convert to julian days counting from the beginning of the year
-    jd_ = to_julian_date(d, m, y)  # julian date of data
-    jd11_ = to_julian_date(1, 1, y)  # julian date 1st Jan
-    jd = jd_ - jd11_ + 1  # days since start of year
-
-    # declination angle + time correction for solar angle
-    d, tc = solar_declination_angle(jd, hh)
-    drad = d * to_radians
-
-    # solar hour angle
-    sha = (hh - 12) * 15 + (lon + 180) / 2 + tc  # solar hour angle
-    sharad = sha * to_radians
-
-    latrad = lat * to_radians
-    lonrad = (lon + 180) / 2 * to_radians
-    accumulationperiod = step  # hours
-
-    # 2*3.1415=day in rad; /24=day hour in rad; *accumulationperiod/2;
-    # = half of the (radiation) accumulation period - SUN IN THE MIDDLE!
-    zhalftimestep = (2 * np.pi) / 24 * accumulationperiod / 2
-    zsolartimestart = sharad - zhalftimestep
-    zsolartimeend = sharad + zhalftimestep
-    ztandec = math.sin(drad) / max((math.cos(drad), 1.0e-12))
-    zcoshouranglesunset = (
-        -ztandec * np.sin(latrad) / np.clip(np.cos(latrad), 1.0e-12, None)
-    )
-    zsindecsinlat = math.sin(drad) * np.sin(latrad)
-    zcosdeccoslat = math.cos(drad) * np.cos(latrad)
-
-    def solar_zenith_angle_average(
-        lonrad,
-        zcoshouranglesunset,
-        zsindecsinlat,
-        zcosdeccoslat,
-        zsolartimestart,
-        zsolartimeend,
-        sha,
-    ):
-        # start and end hour
-        def horizon(val, rrange, lonrad, zsolartimestart, zsolartimeend, maxx):
-            if val < rrange * math.pi:
-                zhouranglestart = zsolartimestart + lonrad - ((rrange - 1.0) * math.pi)
-                zhourangleend = zsolartimeend + lonrad - ((rrange - 1.0) * math.pi)
+    # Gauss-Integration coefficients
+    if integration_order == 3:  # default, good speed and accuracy (3 points)
+        E = np.array([-math.sqrt(3.0 / 5.0), 0.0, math.sqrt(3.0 / 5.0)])
+        W = np.array([5.0 / 9.0, 8.0 / 9.0, 5.0 / 9.0])
+    else:
+        if integration_order == 1:  # fastest, worse accuracy (1 point)
+            E = np.array([0.0])
+            W = np.array([2.0])
+        else:
+            if integration_order == 2:  # faster, less accurate (2 points)
+                E = np.array([-1.0 / math.sqrt(3.0), 1.0 / math.sqrt(3.0)])
+                W = np.array([1.0, 1.0])
             else:
-                if rrange < maxx:
-                    zhouranglestart, zhourangleend = horizon(
-                        val, rrange + 2.0, lonrad, zsolartimestart, zsolartimeend, maxx
+                if integration_order == 4:  # slower, more accurate (4 points)
+                    E = np.array(
+                        [
+                            -math.sqrt(3.0 / 7.0 + 2.0 / 7.0 * math.sqrt(6.0 / 5.0)),
+                            -math.sqrt(3.0 / 7.0 - 2.0 / 7.0 * math.sqrt(6.0 / 5.0)),
+                            math.sqrt(3.0 / 7.0 - 2.0 / 7.0 * math.sqrt(6.0 / 5.0)),
+                            math.sqrt(3.0 / 7.0 + 2.0 / 7.0 * math.sqrt(6.0 / 5.0)),
+                        ]
+                    )
+                    W = np.array(
+                        [
+                            (18 - math.sqrt(30)) / 36,
+                            (18 + math.sqrt(30)) / 36,
+                            (18 + math.sqrt(30)) / 36,
+                            (18 - math.sqrt(30)) / 36,
+                        ]
                     )
                 else:
-                    zhouranglestart = zsolartimestart + lonrad - (maxx + 1.0) * math.pi
-                    zhourangleend = zsolartimeend + lonrad - (maxx + 1.0) * math.pi
+                    print(f"Invalid integration_order {integration_order}")
+                    raise ValueError
 
-            return (zhouranglestart, zhourangleend)
+    assert intervals_per_hour > 0
 
-        # calculating the solar zenith angle
+    nsplits = (tend - tbegin) * intervals_per_hour
 
-        PMU0 = 0
-        if zcoshouranglesunset <= 1:
-            zhouranglestart, zhourangleend = horizon(
-                sha * math.pi / 180 + lonrad,
-                2,
-                lonrad,
-                zsolartimestart,
-                zsolartimeend,
-                maxx,
+    assert nsplits > 0
+
+    time_steps = np.linspace(tbegin, tend, num=nsplits + 1)
+
+    integral = np.zeros_like(lat)
+
+    for s in range(len(time_steps) - 1):
+        ti = time_steps[s]
+        tf = time_steps[s + 1]
+
+        # print(f"Interval {s+1} [{ti}, {tf}]")
+
+        deltat = tf - ti
+        jacob = deltat / 2.0
+
+        w = jacob * W
+        w /= tend - tbegin  # average of integral
+        t = jacob * E
+        t += (tf + ti) / 2.0
+
+        # print(f"w {w}")
+        # print(f"t {t}")
+
+        for n in range(len(w)):
+            cossza = calculate_cos_solar_zenith_angle(
+                lat=lat, lon=lon, y=y, m=m, d=d, h=(h + t[n])
             )
+            integral += w[n] * cossza
 
-            if zcoshouranglesunset >= -1:
-                zhouranglesunset = math.acos(zcoshouranglesunset)
-                if (
-                    zhourangleend <= -zhouranglesunset
-                    or zhouranglestart >= zhouranglesunset
-                ):
-                    PMU0 = 0
-                zhouranglestart = max(
-                    -zhouranglesunset, min(zhouranglestart, zhouranglesunset)
-                )
-                zhourangleend = max(
-                    -zhouranglesunset, min(zhourangleend, zhouranglesunset)
-                )
+    # integral /= (tend - tbegin) # average is above for efficiency
 
-            if (zhourangleend - zhouranglestart) > 1.0e-8:
-                PMU0 = max(
-                    0.0,
-                    zsindecsinlat
-                    + (
-                        zcosdeccoslat
-                        * (math.sin(zhourangleend) - math.sin(zhouranglestart))
-                    )
-                    / (zhourangleend - zhouranglestart),
-                )
-            else:
-                PMU0 = 0.0
-
-        return PMU0
-
-    return np.vectorize(solar_zenith_angle_average)(
-        lonrad,
-        zcoshouranglesunset,
-        zsindecsinlat,
-        zcosdeccoslat,
-        zsolartimestart,
-        zsolartimeend,
-        sha,
-    )
+    return integral
 
 
 def calculate_mean_radiant_temperature(ssrd, ssr, fdir, strd, strr, cossza):
@@ -358,13 +298,14 @@ def calculate_mean_radiant_temperature(ssrd, ssr, fdir, strd, strr, cossza):
     return mrt
 
 
-def calculate_utci(t2_k, va_ms, mrt_k, e_hPa):
+def calculate_utci(t2_k, va_ms, mrt_k, e_hPa=None, td_k=None):
     """
     UTCI
-    :param t2m: (float array) is 2m temperature [K]
-    :param va: (float array) is wind speed at 10 meters [m/s]
-    :param mrt:(float array) is mean radiant temperature [K]
-    :param ehPa: (float array) is water vapour pressure [hPa]
+    :param t2_k: (float array) is 2m temperature [K]
+    :param va_ms: (float array) is wind speed at 10 meters [m/s]
+    :param mrt_k:(float array) is mean radiant temperature [K]
+    :param e_hPa: (float array) is water vapour pressure [hPa]
+    :param td_k: (float array) is 2m dew point temperature [K]
 
     Calculate UTCI with a 6th order polynomial approximation according to:
     Brode, P. et al. Deriving the operational procedure for the
@@ -376,12 +317,21 @@ def calculate_utci(t2_k, va_ms, mrt_k, e_hPa):
     t2 = __wrap(t2_k)
     va = __wrap(va_ms)
     mrt_kw = __wrap(mrt_k)
-    ehPa = __wrap(e_hPa)
 
-    rh = ehPa / 10.0  # rh in kPa
+    if e_hPa is not None:
+        ehPa = __wrap(e_hPa)
+        rh = ehPa / 10.0  # rh in kPa
+    else:
+        if td_k is not None:
+            t2d = __wrap(td_k)
+            rh_pc = calculate_relative_humidity_percent(t2, t2d)
+            ehPa = calculate_saturation_vapour_pressure(t2) * rh_pc / 100.0
+            rh = ehPa / 10.0  # rh in kPa
+        else:
+            raise ValueError("Missing input e_hPa or td_k")
 
-    t2m = kelvin_to_celcius(t2)  # polynomial approx. is in Celsius
-    mrt = kelvin_to_celcius(mrt_kw)  # polynomial approx. is in Celsius
+    t2m = kelvin_to_celsius(t2)  # polynomial approx. is in Celsius
+    mrt = kelvin_to_celsius(mrt_kw)  # polynomial approx. is in Celsius
 
     e_mrt = np.subtract(mrt, t2m)
 
@@ -625,27 +575,32 @@ def calculate_utci(t2_k, va_ms, mrt_k, e_hPa):
 
     # print(f"utci {utci}")
 
-    utci_filtert2m = np.where(t2m >= 70)
-    utci_filtert2m2 = np.where(t2m <= -70)
-    utci_filterva = np.where(17 <= va)
-    utci_filterva2 = np.where(0 >= va)
-    utci_filterrh = np.where(5 < rh)
-    utci_filtere_mrt = np.where(e_mrt >= 100.0)
-    utci_filtere_mrt2 = np.where(e_mrt <= -30)
-
-    utci[utci_filtert2m] = -9999
+    # utci_filtert2m = np.where(t2m >= 70)
+    # utci[utci_filtert2m] = -9999
     # print(f"utci f1 {utci}")
-    utci[utci_filtert2m2] = -9999
+
+    # utci_filtert2m2 = np.where(t2m <= -70)
+    # utci[utci_filtert2m2] = -9999
     # print(f"utci f2 {utci}")
-    utci[utci_filterva] = -9999
+
+    # utci_filterva = np.where(17 <= va)
+    # utci[utci_filterva] = -9999
     # print(f"utci f3 {utci}")
-    utci[utci_filterva2] = -9999
+
+    # utci_filterva2 = np.where(0 >= va)
+    # utci[utci_filterva2] = -9999
     # print(f"utci f4 {utci}")
-    utci[utci_filterrh] = -9999
+
+    # utci_filterrh = np.where(5 < rh)
+    # utci[utci_filterrh] = -9999
     # print(f"utci f5 {utci}")
-    utci[utci_filtere_mrt] = -9999
+
+    # utci_filtere_mrt = np.where(e_mrt >= 100.0)
+    # utci[utci_filtere_mrt] = -9999
     # print(f"utci f6 {utci}")
-    utci[utci_filtere_mrt2] = -9999
+
+    # utci_filtere_mrt2 = np.where(e_mrt <= -30)
+    # utci[utci_filtere_mrt2] = -9999
     # print(f"utci f7 {utci}")
 
     # print(f"utci {utci}")
@@ -668,32 +623,53 @@ def calculate_wbgts(t2m):
     t2m = __wrap(t2m)
     rh = calculate_saturation_vapour_pressure(t2m)
     rh = kPa_to_hPa(rh)
-    t2m = kelvin_to_celcius(t2m)
+    t2m = kelvin_to_celsius(t2m)
     wbgts = 0.567 * t2m + 0.393 * rh + 3.38
     return wbgts
 
 
-def calculate_wbgt(t2m, mrt, va):
+def calculate_wbt(t_c, rh):
     """
-    calculate wet bulb globe temperature
+    calculate wet globe temperature
+    :param t2m: 2m temperature [°C]
+    :param rh: relative humidity percentage[%]
+
+    returns wet bulb temperature [°C]
+    """
+    t_c = __wrap(t_c)
+    rh = __wrap(rh)
+
+    tw = (
+        t_c * np.arctan(0.151977 * np.sqrt(rh + 8.313659))
+        + np.arctan(t_c + rh)
+        - np.arctan(rh - 1.676331)
+        + 0.00391838 * (rh) ** (3 / 2) * np.arctan(0.023101 * rh)
+        - 4.686035
+    )
+    return tw
+
+
+def calculate_bgt(t_k, mrt, va):
+    """
+    calculate globe temperature
     :param t2m: 2m temperature [K]
     :param mrt: mean radiant temperature [K]
     :param va: wind speed at 10 meters [m/s]
 
-    returns wet bulb globe temperature [°C]
+    returns bulb globe temperature [°C]
     """
-    t2m = __wrap(t2m)
+    t_k = __wrap(t_k)
     mrt = __wrap(mrt)
     va = __wrap(va)
 
     f = (1.1e8 * va ** 0.6) / (0.98 * 0.15 ** 0.4)
     a = f / 2
-    b = -f * t2m - mrt ** 4
+    b = -f * t_k - mrt ** 4
     rt1 = 3 ** (1 / 3)
     rt2 = np.sqrt(3) * np.sqrt(27 * a ** 4 - 16 * b ** 3) + 9 * a ** 2
     rt3 = 2 * 2 ** (2 / 3) * b
     a = a.clip(min=0)
-    wbgt_quartic = -1 / 2 * np.sqrt(
+    bgt_quartic = -1 / 2 * np.sqrt(
         rt3 / (rt1 * rt2 ** (1 / 3)) + (2 ** (1 / 3) * rt2 ** (1 / 3)) / 3 ** (2 / 3)
     ) + 1 / 2 * np.sqrt(
         (4 * a)
@@ -704,27 +680,57 @@ def calculate_wbgt(t2m, mrt, va):
         - (2 ** (1 / 3) * rt2 ** (1 / 3)) / 3 ** (2 / 3)
         - rt3 / (rt1 * rt2 ** (1 / 3))
     )
-    wbgt_quartic = kelvin_to_celcius(wbgt_quartic)
-    return wbgt_quartic
+
+    bgt_c = kelvin_to_celsius(bgt_quartic)
+    return bgt_c
 
 
-def calculate_mrt_from_wbgt(t2m, wbgt, va):
+def calculate_wbgt(t_k, mrt, va, td):
+    """
+    calculate wet bulb globe temperature
+    :param t_k: 2m temperature [K]
+    :param mrt: mean radiant temperature [K]
+    :param va: wind speed at 10 meters [m/s]
+    :param td: dew point temperature [°C]
+
+    returns wet bulb globe temperature [°C]
+
+    https://journals.ametsoc.org/view/journals/apme/50/11/jamc-d-11-0143.1.xml
+
+    """
+    t_k = __wrap(t_k)
+    mrt = __wrap(mrt)
+    va = __wrap(va)
+    td = __wrap(td)
+
+    bgt_c = calculate_bgt(t_k, mrt, va)
+
+    rh = calculate_relative_humidity_percent(t_k, td)
+
+    t_c = kelvin_to_celsius(t_k)
+    tw_c = calculate_wbt(t_c, rh)
+
+    wbgt = 0.7 * tw_c + 0.2 * bgt_c + 0.1 * t_c
+    return wbgt
+
+
+def calculate_mrt_from_bgt(t2m, bgt, va):
     """
     calculate mean radiant temperature from wet bulb globe temperature
     :param t2m: 2m temperature [K]
-    :param wbgt: wet bulb globe temperature in kelvin [K]
+    :param bgt: bulb globe temperature in Kelvin [K]
     :param va: wind speed at 10 meters [m/s]
     returns mean radiant temperature [K]
     """
     t2m = __wrap(t2m)
-    wbgt = __wrap(wbgt)
+    bgt = __wrap(bgt)
     va = __wrap(va)
 
     f = (1.1e8 * va ** 0.6) / (0.98 * 0.15 ** 0.4)
-    wbgt4 = wbgt ** 4
-    mrtc = wbgt4 + f * (wbgt - t2m)
+    bgt4 = bgt ** 4
+    mrtc = bgt4 + f * (bgt - t2m)
     mrtc2 = np.sqrt(np.sqrt(mrtc))
-    return kelvin_to_celcius(mrtc2)
+    return kelvin_to_celsius(mrtc2)
 
 
 def calculate_humidex(t2m, td):
@@ -745,7 +751,7 @@ def calculate_humidex(t2m, td):
 
 def calculate_net_effective_temperature(t2m, va, td):
     """
-    Net Effective Temperature used in Hong Kong, Poland and Germany
+    Net - Normal Effective Temperature used in Hong Kong, Poland and Germany
     :param t2m: 2m temperature [K]
     :param td: 2m dew point temperature [K]
     :param rh: Relative Humidity [pa]
@@ -757,7 +763,7 @@ def calculate_net_effective_temperature(t2m, va, td):
     t2m = __wrap(t2m)
     va = __wrap(va)
     rh = __wrap(rh)
-    t2m = kelvin_to_celcius(t2m)
+    t2m = kelvin_to_celsius(t2m)
     rh = kPa_to_hPa(rh)
     ditermeq = 1 / 1.76 + 1.4 * va ** 0.75
     net = 37 - (37 - t2m / 0.68 - 0.0014 * rh + ditermeq) - 0.29 * t2m * (1 - 0.01 * rh)
@@ -779,7 +785,7 @@ def calculate_apparent_temperature(t2m, va, rh=None):
     rh = __wrap(rh)
     va = va * 4.87 / np.log10(67.8 * 10 - 5.42)  # converting to 2m, ~1.2m wind speed
     at = t2m + 0.33 * rh - 0.7 * va - 4
-    at = kelvin_to_celcius(at)
+    at = kelvin_to_celsius(at)
     return at
 
 
@@ -793,7 +799,7 @@ def calculate_wind_chill(t2m, va):
     """
     t2m = __wrap(t2m)
     va = __wrap(va)
-    t2m = kelvin_to_celcius(t2m)
+    t2m = kelvin_to_celsius(t2m)
     va = va * 2.23694  # convert to miles per hour
     windchill = 13.12 + 0.6215 * t2m - 11.37 * va ** 0.16 + 0.3965 + t2m + va ** 0.16
     return windchill
@@ -809,7 +815,7 @@ def calculate_heat_index_simplified(t2m, rh=None):
     t2m = __wrap(t2m)
     if rh is None:
         rh = calculate_saturation_vapour_pressure(t2m)
-    t2m = kelvin_to_celcius(t2m)
+    t2m = kelvin_to_celsius(t2m)
     rh = kPa_to_hPa(rh)
 
     hiarray = [
@@ -848,7 +854,7 @@ def calculate_heat_index_adjusted(t2m, td):
     td = __wrap(td)
 
     rh = calculate_relative_humidity_percent(t2m, td)
-    t2m = kelvin_to_farenheit(t2m)
+    t2m = kelvin_to_fahrenheit(t2m)
 
     hiarray = [
         42.379,
@@ -905,19 +911,19 @@ def calculate_heat_index_adjusted(t2m, td):
         hi[hi_filter1 and hi_filter4 and hi_filter5] - adjustment2
     )
     hi[hi_filter6] = adjustment3
-    hi = farenheit_to_celcius(hi)
+    hi = fahrenheit_to_celsius(hi)
     return hi
 
 
 # Helpers
 
-# convert celcius to kelvin
-def celcius_to_kelvin(tc):
+# convert Celsius to Kelvin
+def celsius_to_kelvin(tc):
     tk = tc + 273.15
     return tk
 
 
-# convert kelvin to celcius
-def kelvin_to_celcius(tk):
+# convert Kelvin to Celsius
+def kelvin_to_celsius(tk):
     tc = tk - 273.15
     return tc
